@@ -33,29 +33,38 @@ macro_rules! attrgen {
             (static_method_of, StaticMethodOf(Span, Ident)),
             (js_namespace, JsNamespace(Span, Ident)),
             (module, Module(Span, String, Span)),
+            (raw_module, RawModule(Span, String, Span)),
+            (inline_js, InlineJs(Span, String, Span)),
             (getter, Getter(Span, Option<Ident>)),
             (setter, Setter(Span, Option<Ident>)),
             (indexing_getter, IndexingGetter(Span)),
             (indexing_setter, IndexingSetter(Span)),
             (indexing_deleter, IndexingDeleter(Span)),
             (structural, Structural(Span)),
-            (final_("final"), Final(Span)),
+            (r#final, Final(Span)),
             (readonly, Readonly(Span)),
             (js_name, JsName(Span, String, Span)),
             (js_class, JsClass(Span, String, Span)),
+            (inspectable, Inspectable(Span)),
+            (is_type_of, IsTypeOf(Span, syn::Expr)),
             (extends, Extends(Span, syn::Path)),
             (vendor_prefix, VendorPrefix(Span, Ident)),
             (variadic, Variadic(Span)),
             (typescript_custom_section, TypescriptCustomSection(Span)),
             (start, Start(Span)),
+            (skip, Skip(Span)),
+
+            // For testing purposes only.
+            (assert_no_shim, AssertNoShim(Span)),
         }
     };
 }
 
 macro_rules! methods {
-    ($(($name:ident $(($other:tt))*, $variant:ident($($contents:tt)*)),)*) => {
+    ($(($name:ident, $variant:ident($($contents:tt)*)),)*) => {
         $(methods!(@method $name, $variant($($contents)*));)*
 
+        #[cfg(feature = "strict-macro")]
         fn check_used(self) -> Result<(), Diagnostic> {
             // Account for the fact this method was called
             ATTRS.with(|state| state.checks.set(state.checks.get() + 1));
@@ -65,9 +74,12 @@ macro_rules! methods {
                 if used.get() {
                     continue
                 }
-                if !cfg!(feature = "strict-macro") {
+                // The check below causes rustc to crash on powerpc64 platforms
+                // with an LLVM error. To avoid this, we instead use #[cfg()]
+                // and duplicate the function below. See #58516 for details.
+                /*if !cfg!(feature = "strict-macro") {
                     continue
-                }
+                }*/
                 let span = match attr {
                     $(BindgenAttr::$variant(span, ..) => span,)*
                 };
@@ -75,6 +87,13 @@ macro_rules! methods {
             }
             Diagnostic::from_vec(errors)
         }
+
+        #[cfg(not(feature = "strict-macro"))]
+        fn check_used(self) -> Result<(), Diagnostic> {
+            // Account for the fact this method was called
+            ATTRS.with(|state| state.checks.set(state.checks.get() + 1));
+            Ok(())
+        }
     };
 
     (@method $name:ident, $variant:ident(Span, String, Span)) => {
@@ -140,7 +159,7 @@ impl BindgenAttrs {
                 None => return Ok(ret),
             };
             let attr = attrs.remove(pos);
-            let mut tts = attr.tts.clone().into_iter();
+            let mut tts = attr.tokens.clone().into_iter();
             let group = match tts.next() {
                 Some(TokenTree::Group(d)) => d,
                 Some(_) => bail_span!(attr, "malformed #[wasm_bindgen] attribute"),
@@ -185,7 +204,7 @@ impl Parse for BindgenAttrs {
 }
 
 macro_rules! gen_bindgen_attr {
-    ($(($method:ident $(($other:tt))*, $($variants:tt)*),)*) => {
+    ($(($method:ident, $($variants:tt)*),)*) => {
         /// The possible attributes in the `#[wasm_bindgen]`.
         #[cfg_attr(feature = "extra-traits", derive(Debug, PartialEq, Eq))]
         pub enum BindgenAttr {
@@ -201,11 +220,13 @@ impl Parse for BindgenAttr {
         let attr: AnyIdent = input.parse()?;
         let attr = attr.0;
         let attr_span = attr.span();
+        let attr_string = attr.to_string();
+        let raw_attr_string = format!("r#{}", attr_string);
 
         macro_rules! parsers {
-            ($(($name:ident $(($other:tt))*, $($contents:tt)*),)*) => {
+            ($(($name:ident, $($contents:tt)*),)*) => {
                 $(
-                    if attr == parsers!(@attrname $name $($other)*) {
+                    if attr_string == stringify!($name) || raw_attr_string == stringify!($name) {
                         parsers!(
                             @parser
                             $($contents)*
@@ -238,6 +259,11 @@ impl Parse for BindgenAttr {
                 return Ok(BindgenAttr::$variant(attr_span, input.parse()?));
             });
 
+            (@parser $variant:ident(Span, syn::Expr)) => ({
+                input.parse::<Token![=]>()?;
+                return Ok(BindgenAttr::$variant(attr_span, input.parse()?));
+            });
+
             (@parser $variant:ident(Span, String, Span)) => ({
                 input.parse::<Token![=]>()?;
                 let (val, span) = match input.parse::<syn::LitStr>() {
@@ -249,9 +275,6 @@ impl Parse for BindgenAttr {
                 };
                 return Ok(BindgenAttr::$variant(attr_span, val, span))
             });
-
-            (@attrname $a:ident $b:tt) => ($b);
-            (@attrname $a:ident) => (stringify!($a));
         }
 
         attrgen!(parsers);
@@ -287,7 +310,7 @@ trait ConvertToAst<Ctx> {
 impl<'a> ConvertToAst<BindgenAttrs> for &'a mut syn::ItemStruct {
     type Target = ast::Struct;
 
-    fn convert(self, opts: BindgenAttrs) -> Result<Self::Target, Diagnostic> {
+    fn convert(self, attrs: BindgenAttrs) -> Result<Self::Target, Diagnostic> {
         if self.generics.params.len() > 0 {
             bail_span!(
                 self.generics,
@@ -296,60 +319,66 @@ impl<'a> ConvertToAst<BindgenAttrs> for &'a mut syn::ItemStruct {
             );
         }
         let mut fields = Vec::new();
-        let js_name = opts
+        let js_name = attrs
             .js_name()
             .map(|s| s.0.to_string())
             .unwrap_or(self.ident.to_string());
-        if let syn::Fields::Named(names) = &mut self.fields {
-            for field in names.named.iter_mut() {
-                match field.vis {
-                    syn::Visibility::Public(..) => {}
-                    _ => continue,
-                }
-                let name = match &field.ident {
-                    Some(n) => n,
-                    None => continue,
-                };
-                let name_str = name.to_string();
-                let getter = shared::struct_field_get(&js_name, &name_str);
-                let setter = shared::struct_field_set(&js_name, &name_str);
-                let opts = BindgenAttrs::find(&mut field.attrs)?;
-                assert_not_variadic(&opts)?;
-                let comments = extract_doc_comments(&field.attrs);
-                fields.push(ast::StructField {
-                    name: name.clone(),
-                    struct_name: self.ident.clone(),
-                    readonly: opts.readonly().is_some(),
-                    ty: field.ty.clone(),
-                    getter: Ident::new(&getter, Span::call_site()),
-                    setter: Ident::new(&setter, Span::call_site()),
-                    comments,
-                });
-                opts.check_used()?;
+        let is_inspectable = attrs.inspectable().is_some();
+        for (i, field) in self.fields.iter_mut().enumerate() {
+            match field.vis {
+                syn::Visibility::Public(..) => {}
+                _ => continue,
             }
+            let (name_str, member) = match &field.ident {
+                Some(ident) => (ident.to_string(), syn::Member::Named(ident.clone())),
+                None => (i.to_string(), syn::Member::Unnamed(i.into())),
+            };
+
+            let attrs = BindgenAttrs::find(&mut field.attrs)?;
+            assert_not_variadic(&attrs)?;
+            if attrs.skip().is_some() {
+                attrs.check_used()?;
+                continue;
+            }
+
+            let comments = extract_doc_comments(&field.attrs);
+            let getter = shared::struct_field_get(&js_name, &name_str);
+            let setter = shared::struct_field_set(&js_name, &name_str);
+
+            fields.push(ast::StructField {
+                name: member,
+                struct_name: self.ident.clone(),
+                readonly: attrs.readonly().is_some(),
+                ty: field.ty.clone(),
+                getter: Ident::new(&getter, Span::call_site()),
+                setter: Ident::new(&setter, Span::call_site()),
+                comments,
+            });
+            attrs.check_used()?;
         }
         let comments: Vec<String> = extract_doc_comments(&self.attrs);
-        opts.check_used()?;
+        attrs.check_used()?;
         Ok(ast::Struct {
             rust_name: self.ident.clone(),
             js_name,
             fields,
             comments,
+            is_inspectable,
         })
     }
 }
 
-impl<'a> ConvertToAst<(BindgenAttrs, &'a Option<String>)> for syn::ForeignItemFn {
+impl<'a> ConvertToAst<(BindgenAttrs, &'a ast::ImportModule)> for syn::ForeignItemFn {
     type Target = ast::ImportKind;
 
     fn convert(
         self,
-        (opts, module): (BindgenAttrs, &'a Option<String>),
+        (opts, module): (BindgenAttrs, &'a ast::ImportModule),
     ) -> Result<Self::Target, Diagnostic> {
         let wasm = function_from_decl(
-            &self.ident,
+            &self.sig.ident,
             &opts,
-            self.decl.clone(),
+            self.sig.clone(),
             self.attrs.clone(),
             self.vis.clone(),
             false,
@@ -371,31 +400,16 @@ impl<'a> ConvertToAst<(BindgenAttrs, &'a Option<String>)> for syn::ForeignItemFn
             wasm.ret.clone()
         };
 
-        let mut operation_kind = ast::OperationKind::Regular;
-        if let Some(g) = opts.getter() {
-            operation_kind = ast::OperationKind::Getter(g.clone());
-        }
-        if let Some(s) = opts.setter() {
-            operation_kind = ast::OperationKind::Setter(s.clone());
-        }
-        if opts.indexing_getter().is_some() {
-            operation_kind = ast::OperationKind::IndexingGetter;
-        }
-        if opts.indexing_setter().is_some() {
-            operation_kind = ast::OperationKind::IndexingSetter;
-        }
-        if opts.indexing_deleter().is_some() {
-            operation_kind = ast::OperationKind::IndexingDeleter;
-        }
+        let operation_kind = operation_kind(&opts);
 
         let kind = if opts.method().is_some() {
             let class = wasm.arguments.get(0).ok_or_else(|| {
                 err_span!(self, "imported methods must have at least one argument")
             })?;
-            let class = match class.ty {
+            let class = match &*class.ty {
                 syn::Type::Reference(syn::TypeReference {
                     mutability: None,
-                    ref elem,
+                    elem,
                     ..
                 }) => &**elem,
                 _ => bail_span!(
@@ -471,7 +485,7 @@ impl<'a> ConvertToAst<(BindgenAttrs, &'a Option<String>)> for syn::ForeignItemFn
                 ast::ImportFunctionKind::Normal => (0, "n"),
                 ast::ImportFunctionKind::Method { ref class, .. } => (1, &class[..]),
             };
-            let data = (ns, &self.ident, module);
+            let data = (ns, &self.sig.ident, module);
             format!(
                 "__wbg_{}_{}",
                 wasm.name
@@ -481,20 +495,22 @@ impl<'a> ConvertToAst<(BindgenAttrs, &'a Option<String>)> for syn::ForeignItemFn
                 ShortHash(data)
             )
         };
-        if let Some(span) = opts.final_() {
+        if let Some(span) = opts.r#final() {
             if opts.structural().is_some() {
                 let msg = "cannot specify both `structural` and `final`";
                 return Err(Diagnostic::span_error(*span, msg));
             }
         }
+        let assert_no_shim = opts.assert_no_shim().is_some();
         let ret = ast::ImportKind::Function(ast::ImportFunction {
             function: wasm,
+            assert_no_shim,
             kind,
             js_ret,
             catch,
             variadic,
-            structural: opts.structural().is_some() || opts.final_().is_none(),
-            rust_name: self.ident.clone(),
+            structural: opts.structural().is_some() || opts.r#final().is_none(),
+            rust_name: self.sig.ident.clone(),
             shim: Ident::new(&shim, Span::call_site()),
             doc_comment: None,
         });
@@ -513,6 +529,7 @@ impl ConvertToAst<BindgenAttrs> for syn::ForeignItemType {
             .js_name()
             .map(|s| s.0)
             .map_or_else(|| self.ident.to_string(), |s| s.to_string());
+        let is_type_of = attrs.is_type_of().cloned();
         let shim = format!("__wbg_instanceof_{}_{}", self.ident, ShortHash(&self.ident));
         let mut extends = Vec::new();
         let mut vendor_prefixes = Vec::new();
@@ -535,6 +552,7 @@ impl ConvertToAst<BindgenAttrs> for syn::ForeignItemType {
             attrs: self.attrs,
             doc_comment: None,
             instanceof_shim: shim,
+            is_type_of,
             rust_name: self.ident,
             js_name,
             extends,
@@ -543,12 +561,12 @@ impl ConvertToAst<BindgenAttrs> for syn::ForeignItemType {
     }
 }
 
-impl<'a> ConvertToAst<(BindgenAttrs, &'a Option<String>)> for syn::ForeignItemStatic {
+impl<'a> ConvertToAst<(BindgenAttrs, &'a ast::ImportModule)> for syn::ForeignItemStatic {
     type Target = ast::ImportKind;
 
     fn convert(
         self,
-        (opts, module): (BindgenAttrs, &'a Option<String>),
+        (opts, module): (BindgenAttrs, &'a ast::ImportModule),
     ) -> Result<Self::Target, Diagnostic> {
         if self.mutability.is_some() {
             bail_span!(self.mutability, "cannot import mutable globals yet")
@@ -584,21 +602,21 @@ impl ConvertToAst<BindgenAttrs> for syn::ItemFn {
             syn::Visibility::Public(_) => {}
             _ => bail_span!(self, "can only #[wasm_bindgen] public functions"),
         }
-        if self.constness.is_some() {
+        if self.sig.constness.is_some() {
             bail_span!(
-                self.constness,
+                self.sig.constness,
                 "can only #[wasm_bindgen] non-const functions"
             );
         }
-        if self.unsafety.is_some() {
-            bail_span!(self.unsafety, "can only #[wasm_bindgen] safe functions");
+        if self.sig.unsafety.is_some() {
+            bail_span!(self.sig.unsafety, "can only #[wasm_bindgen] safe functions");
         }
         assert_not_variadic(&attrs)?;
 
         let ret = function_from_decl(
-            &self.ident,
+            &self.sig.ident,
             &attrs,
-            self.decl,
+            self.sig.clone(),
             self.attrs,
             self.vis,
             false,
@@ -613,25 +631,25 @@ impl ConvertToAst<BindgenAttrs> for syn::ItemFn {
 fn function_from_decl(
     decl_name: &syn::Ident,
     opts: &BindgenAttrs,
-    decl: Box<syn::FnDecl>,
+    sig: syn::Signature,
     attrs: Vec<syn::Attribute>,
     vis: syn::Visibility,
     allow_self: bool,
     self_ty: Option<&Ident>,
 ) -> Result<(ast::Function, Option<ast::MethodSelf>), Diagnostic> {
-    if decl.variadic.is_some() {
-        bail_span!(decl.variadic, "can't #[wasm_bindgen] variadic functions");
+    if sig.variadic.is_some() {
+        bail_span!(sig.variadic, "can't #[wasm_bindgen] variadic functions");
     }
-    if decl.generics.params.len() > 0 {
+    if sig.generics.params.len() > 0 {
         bail_span!(
-            decl.generics,
+            sig.generics,
             "can't #[wasm_bindgen] functions with lifetime or type parameters",
         );
     }
 
-    assert_no_lifetimes(&decl)?;
+    assert_no_lifetimes(&sig)?;
 
-    let syn::FnDecl { inputs, output, .. } = { *decl };
+    let syn::Signature { inputs, output, .. } = sig;
 
     let replace_self = |t: syn::Type| {
         let self_ty = match self_ty {
@@ -657,25 +675,24 @@ fn function_from_decl(
     let arguments = inputs
         .into_iter()
         .filter_map(|arg| match arg {
-            syn::FnArg::Captured(mut c) => {
-                c.ty = replace_self(c.ty);
+            syn::FnArg::Typed(mut c) => {
+                c.ty = Box::new(replace_self(*c.ty));
                 Some(c)
             }
-            syn::FnArg::SelfValue(_) => {
+            syn::FnArg::Receiver(r) => {
+                if !allow_self {
+                    panic!("arguments cannot be `self`")
+                }
                 assert!(method_self.is_none());
-                method_self = Some(ast::MethodSelf::ByValue);
-                None
-            }
-            syn::FnArg::SelfRef(ref a) if allow_self => {
-                assert!(method_self.is_none());
-                if a.mutability.is_some() {
+                if r.reference.is_none() {
+                    method_self = Some(ast::MethodSelf::ByValue);
+                } else if r.mutability.is_some() {
                     method_self = Some(ast::MethodSelf::RefMutable);
                 } else {
                     method_self = Some(ast::MethodSelf::RefShared);
                 }
                 None
             }
-            _ => panic!("arguments cannot be `self` or ignored"),
         })
         .collect::<Vec<_>>();
 
@@ -684,18 +701,22 @@ fn function_from_decl(
         syn::ReturnType::Type(_, ty) => Some(replace_self(*ty)),
     };
 
-    let js_name = opts.js_name();
+    let (name, name_span, renamed_via_js_name) =
+        if let Some((js_name, js_name_span)) = opts.js_name() {
+            (js_name.to_string(), js_name_span, true)
+        } else {
+            (decl_name.to_string(), decl_name.span(), false)
+        };
     Ok((
         ast::Function {
-            name: js_name
-                .map(|s| s.0.to_string())
-                .unwrap_or(decl_name.to_string()),
-            name_span: js_name.map(|s| s.1).unwrap_or(decl_name.span()),
-            renamed_via_js_name: js_name.is_some(),
             arguments,
+            name_span,
+            name,
+            renamed_via_js_name,
             ret,
-            rust_vis: vis,
             rust_attrs: attrs,
+            rust_vis: vis,
+            r#async: sig.asyncness.is_some(),
         },
         method_self,
     ))
@@ -721,8 +742,8 @@ impl<'a> MacroParse<(Option<BindgenAttrs>, &'a mut TokenStream)> for syn::Item {
                     .attrs
                     .iter()
                     .enumerate()
-                    .filter_map(|(i, m)| m.interpret_meta().map(|m| (i, m)))
-                    .find(|&(_, ref m)| m.name() == "no_mangle");
+                    .filter_map(|(i, m)| m.parse_meta().ok().map(|m| (i, m)))
+                    .find(|(_, m)| m.path().is_ident("no_mangle"));
                 match no_mangle {
                     Some((i, _)) => {
                         f.attrs.remove(i);
@@ -733,22 +754,28 @@ impl<'a> MacroParse<(Option<BindgenAttrs>, &'a mut TokenStream)> for syn::Item {
                 f.to_tokens(tokens);
                 let opts = opts.unwrap_or_default();
                 if opts.start().is_some() {
-                    if f.decl.generics.params.len() > 0 {
-                        bail_span!(&f.decl.generics, "the start function cannot have generics",);
+                    if f.sig.generics.params.len() > 0 {
+                        bail_span!(&f.sig.generics, "the start function cannot have generics",);
                     }
-                    if f.decl.inputs.len() > 0 {
-                        bail_span!(&f.decl.inputs, "the start function cannot have arguments",);
+                    if f.sig.inputs.len() > 0 {
+                        bail_span!(&f.sig.inputs, "the start function cannot have arguments",);
                     }
                 }
+                let method_kind = ast::MethodKind::Operation(ast::Operation {
+                    is_static: true,
+                    kind: operation_kind(&opts),
+                });
+                let rust_name = f.sig.ident.clone();
+                let start = opts.start().is_some();
                 program.exports.push(ast::Export {
-                    rust_class: None,
-                    js_class: None,
-                    method_self: None,
-                    is_constructor: false,
                     comments,
-                    rust_name: f.ident.clone(),
-                    start: opts.start().is_some(),
                     function: f.convert(opts)?,
+                    js_class: None,
+                    method_kind,
+                    method_self: None,
+                    rust_class: None,
+                    rust_name,
+                    start,
                 });
             }
             syn::Item::Struct(mut s) => {
@@ -796,7 +823,11 @@ impl<'a> MacroParse<(Option<BindgenAttrs>, &'a mut TokenStream)> for syn::Item {
 }
 
 impl<'a> MacroParse<BindgenAttrs> for &'a mut syn::ItemImpl {
-    fn macro_parse(self, program: &mut ast::Program, opts: BindgenAttrs) -> Result<(), Diagnostic> {
+    fn macro_parse(
+        self,
+        _program: &mut ast::Program,
+        opts: BindgenAttrs,
+    ) -> Result<(), Diagnostic> {
         if self.defaultness.is_some() {
             bail_span!(
                 self.defaultness,
@@ -830,7 +861,7 @@ impl<'a> MacroParse<BindgenAttrs> for &'a mut syn::ItemImpl {
         };
         let mut errors = Vec::new();
         for item in self.items.iter_mut() {
-            if let Err(e) = (&name, item).macro_parse(program, &opts) {
+            if let Err(e) = prepare_for_impl_recursion(item, &name, &opts) {
                 errors.push(e);
             }
         }
@@ -840,77 +871,111 @@ impl<'a> MacroParse<BindgenAttrs> for &'a mut syn::ItemImpl {
     }
 }
 
-impl<'a, 'b> MacroParse<&'a BindgenAttrs> for (&'a Ident, &'b mut syn::ImplItem) {
+// Prepare for recursion into an `impl` block. Here we want to attach an
+// internal attribute, `__wasm_bindgen_class_marker`, with any metadata we need
+// to pass from the impl to the impl item. Recursive macro expansion will then
+// expand the `__wasm_bindgen_class_marker` attribute.
+//
+// Note that we currently do this because inner items may have things like cfgs
+// on them, so we want to expand the impl first, let the insides get cfg'd, and
+// then go for the rest.
+fn prepare_for_impl_recursion(
+    item: &mut syn::ImplItem,
+    class: &Ident,
+    impl_opts: &BindgenAttrs,
+) -> Result<(), Diagnostic> {
+    let method = match item {
+        syn::ImplItem::Method(m) => m,
+        syn::ImplItem::Const(_) => {
+            bail_span!(
+                &*item,
+                "const definitions aren't supported with #[wasm_bindgen]"
+            );
+        }
+        syn::ImplItem::Type(_) => bail_span!(
+            &*item,
+            "type definitions in impls aren't supported with #[wasm_bindgen]"
+        ),
+        syn::ImplItem::Macro(_) => {
+            // In theory we want to allow this, but we have no way of expanding
+            // the macro and then placing our magical attributes on the expanded
+            // functions. As a result, just disallow it for now to hopefully
+            // ward off buggy results from this macro.
+            bail_span!(&*item, "macros in impls aren't supported");
+        }
+        syn::ImplItem::Verbatim(_) => panic!("unparsed impl item?"),
+        other => bail_span!(other, "failed to parse this item as a known item"),
+    };
+
+    let js_class = impl_opts
+        .js_class()
+        .map(|s| s.0.to_string())
+        .unwrap_or(class.to_string());
+
+    method.attrs.insert(
+        0,
+        syn::Attribute {
+            pound_token: Default::default(),
+            style: syn::AttrStyle::Outer,
+            bracket_token: Default::default(),
+            path: syn::parse_quote! { wasm_bindgen::prelude::__wasm_bindgen_class_marker },
+            tokens: quote::quote! { (#class = #js_class) }.into(),
+        },
+    );
+
+    Ok(())
+}
+
+impl<'a, 'b> MacroParse<(&'a Ident, &'a str)> for &'b mut syn::ImplItemMethod {
     fn macro_parse(
         self,
         program: &mut ast::Program,
-        impl_opts: &'a BindgenAttrs,
+        (class, js_class): (&'a Ident, &'a str),
     ) -> Result<(), Diagnostic> {
-        let (class, item) = self;
-        let method = match item {
-            syn::ImplItem::Method(ref mut m) => m,
-            syn::ImplItem::Const(_) => {
-                bail_span!(
-                    &*item,
-                    "const definitions aren't supported with #[wasm_bindgen]"
-                );
-            }
-            syn::ImplItem::Type(_) => bail_span!(
-                &*item,
-                "type definitions in impls aren't supported with #[wasm_bindgen]"
-            ),
-            syn::ImplItem::Existential(_) => bail_span!(
-                &*item,
-                "existentials in impls aren't supported with #[wasm_bindgen]"
-            ),
-            syn::ImplItem::Macro(_) => {
-                bail_span!(&*item, "macros in impls aren't supported");
-            }
-            syn::ImplItem::Verbatim(_) => panic!("unparsed impl item?"),
-        };
-        match method.vis {
+        match self.vis {
             syn::Visibility::Public(_) => {}
             _ => return Ok(()),
         }
-        if method.defaultness.is_some() {
+        if self.defaultness.is_some() {
             panic!("default methods are not supported");
         }
-        if method.sig.constness.is_some() {
+        if self.sig.constness.is_some() {
             bail_span!(
-                method.sig.constness,
+                self.sig.constness,
                 "can only #[wasm_bindgen] non-const functions",
             );
         }
-        if method.sig.unsafety.is_some() {
-            bail_span!(method.sig.unsafety, "can only bindgen safe functions",);
+        if self.sig.unsafety.is_some() {
+            bail_span!(self.sig.unsafety, "can only bindgen safe functions",);
         }
 
-        let opts = BindgenAttrs::find(&mut method.attrs)?;
-        let comments = extract_doc_comments(&method.attrs);
-        let is_constructor = opts.constructor().is_some();
+        let opts = BindgenAttrs::find(&mut self.attrs)?;
+        let comments = extract_doc_comments(&self.attrs);
         let (function, method_self) = function_from_decl(
-            &method.sig.ident,
+            &self.sig.ident,
             &opts,
-            Box::new(method.sig.decl.clone()),
-            method.attrs.clone(),
-            method.vis.clone(),
+            self.sig.clone(),
+            self.attrs.clone(),
+            self.vis.clone(),
             true,
             Some(class),
         )?;
-        let js_class = impl_opts
-            .js_class()
-            .map(|s| s.0.to_string())
-            .unwrap_or(class.to_string());
-
+        let method_kind = if opts.constructor().is_some() {
+            ast::MethodKind::Constructor
+        } else {
+            let is_static = method_self.is_none();
+            let kind = operation_kind(&opts);
+            ast::MethodKind::Operation(ast::Operation { is_static, kind })
+        };
         program.exports.push(ast::Export {
-            rust_class: Some(class.clone()),
-            js_class: Some(js_class),
-            method_self,
-            is_constructor,
-            function,
             comments,
+            function,
+            js_class: Some(js_class.to_string()),
+            method_kind,
+            method_self,
+            rust_class: Some(class.clone()),
+            rust_name: self.sig.ident.clone(),
             start: false,
-            rust_name: method.sig.ident.clone(),
         });
         opts.check_used()?;
         Ok(())
@@ -924,6 +989,12 @@ impl MacroParse<()> for syn::ItemEnum {
             _ => bail_span!(self, "only public enums are allowed with #[wasm_bindgen]"),
         }
 
+        if self.variants.len() == 0 {
+            bail_span!(self, "cannot export empty enums to JS");
+        }
+
+        let has_discriminant = self.variants[0].discriminant.is_some();
+
         let variants = self
             .variants
             .iter()
@@ -933,25 +1004,36 @@ impl MacroParse<()> for syn::ItemEnum {
                     syn::Fields::Unit => (),
                     _ => bail_span!(v.fields, "only C-Style enums allowed with #[wasm_bindgen]"),
                 }
-                let value = match v.discriminant {
+
+                // Require that everything either has a discriminant or doesn't.
+                // We don't really want to get in the business of emulating how
+                // rustc assigns values to enums.
+                if v.discriminant.is_some() != has_discriminant {
+                    bail_span!(
+                        v,
+                        "must either annotate discriminant of all variants or none"
+                    );
+                }
+
+                let value = match &v.discriminant {
                     Some((
                         _,
                         syn::Expr::Lit(syn::ExprLit {
                             attrs: _,
-                            lit: syn::Lit::Int(ref int_lit),
+                            lit: syn::Lit::Int(int_lit),
                         }),
-                    )) => {
-                        if int_lit.value() > <u32>::max_value() as u64 {
+                    )) => match int_lit.base10_digits().parse::<u32>() {
+                        Ok(v) => v,
+                        Err(_) => {
                             bail_span!(
                                 int_lit,
                                 "enums with #[wasm_bindgen] can only support \
                                  numbers that can be represented as u32"
                             );
                         }
-                        int_lit.value() as u32
-                    }
+                    },
                     None => i as u32,
-                    Some((_, ref expr)) => bail_span!(
+                    Some((_, expr)) => bail_span!(
                         expr,
                         "enums with #[wasm_bidngen] may only have \
                          number literal values",
@@ -963,12 +1045,31 @@ impl MacroParse<()> for syn::ItemEnum {
                     value,
                 })
             })
-            .collect::<Result<_, Diagnostic>>()?;
+            .collect::<Result<Vec<_>, Diagnostic>>()?;
+
+        let mut values = variants.iter().map(|v| v.value).collect::<Vec<_>>();
+        values.sort();
+        let hole = values
+            .windows(2)
+            .filter_map(|window| {
+                if window[0] + 1 != window[1] {
+                    Some(window[0] + 1)
+                } else {
+                    None
+                }
+            })
+            .next()
+            .unwrap_or(*values.last().unwrap() + 1);
+        for value in values {
+            assert!(hole != value);
+        }
+
         let comments = extract_doc_comments(&self.attrs);
         program.enums.push(ast::Enum {
             name: self.ident,
             variants,
             comments,
+            hole,
         });
         Ok(())
     }
@@ -1012,8 +1113,31 @@ impl MacroParse<BindgenAttrs> for syn::ItemForeignMod {
                 ));
             }
         }
-        for mut item in self.items.into_iter() {
-            if let Err(e) = item.macro_parse(program, &opts) {
+        let module = if let Some((name, span)) = opts.module() {
+            if opts.inline_js().is_some() {
+                let msg = "cannot specify both `module` and `inline_js`";
+                errors.push(Diagnostic::span_error(span, msg));
+            }
+            if opts.raw_module().is_some() {
+                let msg = "cannot specify both `module` and `raw_module`";
+                errors.push(Diagnostic::span_error(span, msg));
+            }
+            ast::ImportModule::Named(name.to_string(), span)
+        } else if let Some((name, span)) = opts.raw_module() {
+            if opts.inline_js().is_some() {
+                let msg = "cannot specify both `raw_module` and `inline_js`";
+                errors.push(Diagnostic::span_error(span, msg));
+            }
+            ast::ImportModule::RawNamed(name.to_string(), span)
+        } else if let Some((js, span)) = opts.inline_js() {
+            let i = program.inline_js.len();
+            program.inline_js.push(js.to_string());
+            ast::ImportModule::Inline(i, span)
+        } else {
+            ast::ImportModule::None
+        };
+        for item in self.items.into_iter() {
+            if let Err(e) = item.macro_parse(program, module.clone()) {
                 errors.push(e);
             }
         }
@@ -1023,11 +1147,11 @@ impl MacroParse<BindgenAttrs> for syn::ItemForeignMod {
     }
 }
 
-impl<'a> MacroParse<&'a BindgenAttrs> for syn::ForeignItem {
+impl MacroParse<ast::ImportModule> for syn::ForeignItem {
     fn macro_parse(
         mut self,
         program: &mut ast::Program,
-        opts: &'a BindgenAttrs,
+        module: ast::ImportModule,
     ) -> Result<(), Diagnostic> {
         let item_opts = {
             let attrs = match self {
@@ -1038,11 +1162,7 @@ impl<'a> MacroParse<&'a BindgenAttrs> for syn::ForeignItem {
             };
             BindgenAttrs::find(attrs)?
         };
-        let module = item_opts
-            .module()
-            .or(opts.module())
-            .map(|s| s.0.to_string());
-        let js_namespace = item_opts.js_namespace().or(opts.js_namespace()).cloned();
+        let js_namespace = item_opts.js_namespace().cloned();
         let kind = match self {
             syn::ForeignItem::Fn(f) => f.convert((item_opts, &module))?,
             syn::ForeignItem::Type(t) => t.convert(item_opts)?,
@@ -1076,8 +1196,7 @@ fn extract_first_ty_param(ty: Option<&syn::Type>) -> Result<Option<syn::Type>, D
     let seg = path
         .segments
         .last()
-        .ok_or_else(|| err_span!(t, "must have at least one segment"))?
-        .into_value();
+        .ok_or_else(|| err_span!(t, "must have at least one segment"))?;
     let generics = match seg.arguments {
         syn::PathArguments::AngleBracketed(ref t) => t,
         _ => bail_span!(t, "must be Result<...>"),
@@ -1085,14 +1204,13 @@ fn extract_first_ty_param(ty: Option<&syn::Type>) -> Result<Option<syn::Type>, D
     let generic = generics
         .args
         .first()
-        .ok_or_else(|| err_span!(t, "must have at least one generic parameter"))?
-        .into_value();
+        .ok_or_else(|| err_span!(t, "must have at least one generic parameter"))?;
     let ty = match generic {
         syn::GenericArgument::Type(t) => t,
         other => bail_span!(other, "must be a type parameter"),
     };
-    match *ty {
-        syn::Type::Tuple(ref t) if t.elems.len() == 0 => return Ok(None),
+    match ty {
+        syn::Type::Tuple(t) if t.elems.len() == 0 => return Ok(None),
         _ => {}
     }
     Ok(Some(ty.clone()))
@@ -1108,7 +1226,7 @@ fn extract_doc_comments(attrs: &[syn::Attribute]) -> Vec<String> {
             if a.path.segments.iter().any(|s| s.ident.to_string() == "doc") {
                 Some(
                     // We want to filter out any Puncts so just grab the Literals
-                    a.tts.clone().into_iter().filter_map(|t| match t {
+                    a.tokens.clone().into_iter().filter_map(|t| match t {
                         TokenTree::Literal(lit) => {
                             // this will always return the quoted string, we deal with
                             // that in the cli when we read in the comments
@@ -1129,7 +1247,7 @@ fn extract_doc_comments(attrs: &[syn::Attribute]) -> Vec<String> {
 }
 
 /// Check there are no lifetimes on the function.
-fn assert_no_lifetimes(decl: &syn::FnDecl) -> Result<(), Diagnostic> {
+fn assert_no_lifetimes(sig: &syn::Signature) -> Result<(), Diagnostic> {
     struct Walk {
         diagnostics: Vec<Diagnostic>,
     }
@@ -1146,7 +1264,7 @@ fn assert_no_lifetimes(decl: &syn::FnDecl) -> Result<(), Diagnostic> {
     let mut walk = Walk {
         diagnostics: Vec::new(),
     };
-    syn::visit::Visit::visit_fn_decl(&mut walk, decl);
+    syn::visit::Visit::visit_signature(&mut walk, sig);
     Diagnostic::from_vec(walk.diagnostics)
 }
 
@@ -1188,3 +1306,23 @@ pub fn assert_all_attrs_checked() {
         assert_eq!(state.parsed.get(), state.checks.get());
     })
 }
+
+fn operation_kind(opts: &BindgenAttrs) -> ast::OperationKind {
+    let mut operation_kind = ast::OperationKind::Regular;
+    if let Some(g) = opts.getter() {
+        operation_kind = ast::OperationKind::Getter(g.clone());
+    }
+    if let Some(s) = opts.setter() {
+        operation_kind = ast::OperationKind::Setter(s.clone());
+    }
+    if opts.indexing_getter().is_some() {
+        operation_kind = ast::OperationKind::IndexingGetter;
+    }
+    if opts.indexing_setter().is_some() {
+        operation_kind = ast::OperationKind::IndexingSetter;
+    }
+    if opts.indexing_deleter().is_some() {
+        operation_kind = ast::OperationKind::IndexingDeleter;
+    }
+    operation_kind
+}
